//*****************************************************************************
// File:        Edge.pas
//
// Classe:      TEdge
//
// Description: TEdge represents one edge in the mesh in 3d. It is formed by an
//              origine and a destination, plus a pointer to one tetra that has
//              this edge.
//
//*****************************************************************************
// Last version: 25-06-2002
//*****************************************************************************

unit Edge;

{$IFDEF DEBUG}
  {$ASSERTIONS ON}
{$ELSEIF True}
  {$ASSERTIONS OFF}
{$IFEND}

interface

uses
  Point,
  Tetra;

type
  TEdge = class
  private
    m_org:      TPoint;
    m_dest:     TPoint;
    m_adjTetra: TTetra; //-- pointer to a tetra that has the edge - only one is ok
    procedure   TestInvariants;
  public
    constructor Create; overload;
    constructor Create(org, dest: TPoint; te: TTetra); overload;
    function    GetOrg : TPoint;
    function    GetDest : TPoint;
    function    GetAdjTetra : TTetra;
    procedure   SetTetra(te: TTetra);
    procedure   Reset(const org, dest: TPoint; const te: TTetra);
    function    Compare(const e: TEdge) : boolean;
    function    HasPoint(const pt: TPoint) : boolean;
  end;


implementation


{ TEdge }

constructor TEdge.Create;
begin
  m_org := nil;
  m_dest := nil;
  m_adjTetra := nil;
end;


constructor TEdge.Create(org, dest: TPoint; te: TTetra);
begin
//  Assert( (org <> nil) and (dest <> nil) and (te <> nil) );
  Assert( (org <> nil) and (dest <> nil) );
  m_org := org;
  m_dest := dest;
  m_adjTetra := te;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Determine if 2 edges are the same by comparing their origin and
//              their destination.  The orientation is NOT a factor with this
//              function (edge a-b == edge b-a); and the tetra connected is not
//              a factor either.
//
// Input:       the edge
//
// Output:      TRUE  -> edge e is the same as self
//              FALSE -> edge e is different
//******************************************************************************
function TEdge.Compare(const e: TEdge): boolean;
begin
  if ( ((self.m_org = e.m_org) or (self.m_org = e.m_dest)) and
       ((self.m_dest = e.m_dest) or (self.m_dest = e.m_org)) ) then
  begin
    Result := true;
  end
  else
  begin
    Result := false;
  end;
end;


function TEdge.GetAdjTetra: TTetra;
begin
  Result := m_adjTetra;
end;


function TEdge.GetDest: TPoint;
begin
  Result := m_dest;
end;


function TEdge.GetOrg: TPoint;
begin
  Result := m_org;
end;


procedure TEdge.Reset(const org, dest: TPoint; const te: TTetra);
begin
//  Assert( (org <> nil) and (dest <> nil) and (te <> nil) );
  Assert( (org <> nil) and (dest <> nil) );
  m_org := org;
  m_dest := dest;
  m_adjTetra := te;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


procedure TEdge.TestInvariants;
begin
//  Assert(m_adjTetra <> nil);
  Assert(m_dest <> nil);
  Assert(m_org <> nil);
end;


procedure TEdge.SetTetra(te: TTetra);
begin
  m_adjTetra := te;
end;


function TEdge.HasPoint(const pt: TPoint): boolean;
begin
  if ( (m_org = pt) or (m_dest = pt) ) then
    Result := true
  else
    Result := false;
end;

end.
