//*****************************************************************************
// File:        InOut.pas
//
// Classe:      TMesh_IO
//
// Description: TMesh_IO is used to read or write to/from an ASCII file the
//              points or the topology of the mesh.
//
// Note:
//
//*****************************************************************************
// Last version: 25-06-2002
//*****************************************************************************

unit InOut;

interface

uses
  SysUtils,
  Mesh,
  Tetra,
  Math,
  Point,
  Classes;

type
  TMesh_IO = class
  private
    m_mesh: TMesh;
    procedure   SaveCurve(times: array of double; nb: integer);
  public
    constructor Create(mesh: TMesh);
    procedure   SavePoints(var f: TextFile);
    procedure   SaveTopology(var f: TextFile);
    procedure   ReadPoints(var f: TextFile);
  end;

implementation

uses
  MainFrm;

{ TMesh_IO }

constructor TMesh_IO.Create(mesh: TMesh);
begin
  m_mesh := mesh;
end;


//******************************************************************************
// Description: Read the points stored in a text file (*.pts)
//
// Input:       (the TextFile, by reference)
//
// Output:      the points are read and added to the current mesh.
//
// Format of the *.pts file:
//    first line: 0 (no attributes) OR 1 (attributes)
//    second line: number of points in the file
//    other lines: each line is a point (with or w/o attributes)
//******************************************************************************
procedure TMesh_IO.ReadPoints(var f: TextFile);
var
  att: integer;
  nb: integer;
  i, j: integer;
  coord: array[0..3] of double;
  temp: double;
  times: array of double;
  startTime, endTime: double;
begin
  read(f, att);
  read(f, nb);
  Form1.ProgressBar1.Min := 0;
  Form1.ProgressBar1.Max := nb;
  Form1.ProgressBar1.Step := 1;
  Form1.ProgressBar1.Position := 0;
  SetLength(times, Floor(nb / 500));
  startTime := Form1.StampTime;
  if (att = 0) then //-- file without attributes
  begin
    for i := 1 to nb do
    begin
      for j := 0 to 2 do
      begin
        read(f, temp);
        coord[j] := temp;
      end;
      m_mesh.InsertPoint(coord[0], coord[1], coord[2]);
      Form1.ProgressBar1.StepIt;
    end;
  end
  else  //-- file with attributes
  begin
    for i := 1 to nb do
    begin
      for j := 0 to 3 do
      begin
        read(f, temp);
        coord[j] := temp;
      end;
      m_mesh.InsertPoint(coord[0], coord[1], coord[2], coord[3]);
      Form1.ProgressBar1.StepIt;
      if (i mod 500 = 0) then
      begin
        endTime := Form1.StampTime;
        times[Floor(i/500) - 1] := (endTime - startTime);
      end;
    end;
  end;
  self.SaveCurve(times, nb);
  Form1.ProgressBar1.Position := 0;
end;


procedure TMesh_IO.SaveCurve(times: array of double; nb: integer);
var
  f1: TextFile;
  i: integer;
begin
  AssignFile(f1, 'd:\temp\curve.txt');
  Rewrite(f1);
  try
    Write(f1, FloatToStrF(0, fffixed, 5, 0) + #9);
    WriteLn(f1, FloatToStrF(0, fffixed, 10, 3));
    for i := 0 to (Floor(nb/500) - 1) do
    begin
      Write(f1, FloatToStrF((i*500) + 500, fffixed, 5, 0) + #9);
      WriteLn(f1, FloatToStrF(times[i], fffixed, 10, 3));
    end;
  finally
    CloseFile(f1);
  end;
end;


//******************************************************************************
// Description: Save the point in a text file
//
// Input:       (the TextFile, by reference)
//
// Output:      the TextFile is 'updated' with the points and their attributes
//              one line per point: x y z att
//
// Remarks:     The 4 points of the BigTetra are not saved.
//******************************************************************************
procedure TMesh_IO.SavePoints(var f: TextFile);
var
  i: integer;
  pt: TPoint;
  lstPts: TList;
begin
  lstPts := m_mesh.GetListPts;
  WriteLn(f, '1');
  WriteLn(f, IntToStr(lstPts.Count - 4));
  for i := 4 to (lstPts.Count - 1) do
  begin
    pt := lstPts[i];
    Write(f, FloatToStrF(pt.x, fffixed, 15, 15) + #9);
    Write(f, FloatToStrF(pt.y, fffixed, 15, 15) + #9);
    Write(f, FloatToStrF(pt.z, fffixed, 15, 15) + #9);
    WriteLn(f, FloatToStrF(pt.data, fffixed, 5, 5));
  end;
end;


procedure TMesh_IO.SaveTopology(var f: TextFile);
var
  i: integer;
  j: integer;
  t: TTetra;
  adj: TTetra;
  pt: TPoint;
  lstTetra: TList;
begin
  lstTetra := m_mesh.GetListTetra;
  for i := 0 to (lstTetra.Count - 1) do
  begin
    t := lstTetra[i];
    Write(f, IntToStr(t.m_no) + '   *   ');
    //-- TP
    for j := 1 to 4 do
    begin
      adj := t.GetT(j);
      if (adj = nil) then
      begin
        Write(f, '-');
      end
      else
      begin
        Write(f, IntToStr(adj.m_no));
      end;
      Write(f, #9);
    end;
    Write(f, '*' + #9);
    //-- TT
    for j := 1 to 4 do
    begin
      pt := t.GetP(j);
      Write(f, IntToStr(pt.m_no) + #9);
    end;
    WriteLn(f, #13);
  end;
end;



end.
