//*****************************************************************************
// File:        Isosurface.pas
//
// Classe:      TIsosurface
//
// Description: TIsosurface is used to compute the isosurface of a data set in
//              3d.
//
// Note:
//
//*****************************************************************************
// Last version: 25-06-2002
//*****************************************************************************

unit Isosurface;

interface

uses
  SysUtils,
  Mesh,
  Tetra,
  Point,
  Vector,
  Dialogs,
  Triangle,
  Classes,
  GeomTools;


type
  TIsosurface = class
  private
    m_mesh:     TMesh;
    procedure   Intersect_3(pt1, pt2, pt3: TPoint; lstTr: TList; te: TTetra;
                            data: double);
    procedure   Intersect_3_equal(pt1, pt2, pt3: TPoint; lstTr: TList;
                                  te: TTetra; data: double);
    procedure   Intersect_4(pt1, pt2, pt3, pt4: TPoint; lstTr: TList;
                            te: TTetra; data: double);
    function    FindOverData(te: TTetra; data: double) : TPoint;
    function    FindUnderData(te: TTetra; data: double) : TPoint;
    function    ComputeSurface(data: double) : TList;
  public
    constructor Create(mesh: TMesh);
    procedure   DrawIsosurface(data: double);
  end;


implementation


{ TIsosurface }

uses
  MainFrm;


constructor TIsosurface.Create(mesh: TMesh);
begin
  m_mesh := mesh;
end;


//******************************************************************************
// Description: Compute each triangle for building an isosurface
//
// Input:       data: the value of the surface to draw
//
// Output:      list of triangles (TTriangle)
//******************************************************************************
function TIsosurface.ComputeSurface(data: double) : TList;
var
  i: integer;
  j: integer;
  k: integer;
  l: integer;
  lstTe: TList;
  te: TTetra;
  inter: array[0..3, 0..2] of double;
  num: integer;
  lamda: double;
  a: TPoint;
  b: TPoint;
  lstTr: TList;
  noEqual: integer;
  equal: array[1..4] of TPoint;
  pt1: TPoint;
  pt2: TPoint;
  pt3: TPoint;
  pt4: TPoint;
begin
  pt1 := TPoint.Create;
  pt2 := TPoint.Create;
  pt3 := TPoint.Create;
  pt4 := TPoint.Create;
  lstTe := m_mesh.GetListTetra;
  lstTr := TList.Create;
  num := 0;
  for i := 0 to (lstTe.Count - 1) do  //-- for each tetrahedron in the mesh
  begin
    te := lstTe[i];
    if (m_mesh.IsOutsideCH(te) = false) then
    begin
      //-- compute the intersection between the surface and each edge of the tetra
      for j := 1 to 3 do
      begin
        a := te.GetP(j);
        for k := j + 1 to 4 do
        begin
          b := te.GetP(k);
          //-- ignore a tetra that contains one or more vertices that are equal
          //-- to the value of the isosurface.  An adjacent tetra will manage
          //-- those points.
          if not( (a.data = b.data) or (a.data = data) or (b.data = data) ) then
          begin
            if ( ((data >= a.data) and (data <= b.data)) or
                 ((data <= a.data) and (data >= b.data)) ) then
            begin
              lamda := ((data - a.data) / (b.data - a.data));
              inter[num][0] := ((1 - lamda) * a.x) + (lamda * b.x);
              inter[num][1] := ((1 - lamda) * a.y) + (lamda * b.y);
              inter[num][2] := ((1 - lamda) * a.z) + (lamda * b.z);
              Inc(num);
            end;
          end;
        end;
      end;
      //-- determine if the data surface intersects 3 or 4 edges of the tetra;
      //-- if it's less, it means that one or more vertices are equal to the
      //-- isosurface (degenerate cases)
      if (num = 3) then
      begin
        pt1.SetCoord(inter[0][0], inter[0][1], inter[0][2]);
        pt2.SetCoord(inter[1][0], inter[1][1], inter[1][2]);
        pt3.SetCoord(inter[2][0], inter[2][1], inter[2][2]);
        Intersect_3(pt1, pt2, pt3, lstTr, te, data);
      end
      else if (num = 4) then
      begin
        pt1.SetCoord(inter[0][0], inter[0][1], inter[0][2]);
        pt2.SetCoord(inter[1][0], inter[1][1], inter[1][2]);
        pt3.SetCoord(inter[2][0], inter[2][1], inter[2][2]);
        pt4.SetCoord(inter[3][0], inter[3][1], inter[3][2]);
        Intersect_4(pt1, pt2, pt3, pt4, lstTr, te, data);
      end
      else  //-- so (num = 0) or (num = 1) or (num = 2)
      begin
        //-- count the number of vertices that are equal to the isosurface
        noEqual := 0;
        for l := 1 to 4 do
        begin
          if (te.GetP(l).data = data) then
          begin
            equal[noEqual + 1] := te.GetP(l);
            Inc(noEqual);
          end;
        end;
        //-- 3 points of the tetra are equal to the surface
        if ( (num = 0) and (noEqual = 3) ) then
        begin
          Intersect_3_equal(equal[1], equal[2], equal[3], lstTr, te, data);
        end
        //-- two vertices are equal to the isosurface and there is 1 intersection
        else if ( (num = 1) and (noEqual = 2) ) then
        begin
          pt1.SetCoord(inter[0][0], inter[0][1], inter[0][2]);
          Intersect_3(pt1, equal[1], equal[2], lstTr, te, data);
        end
        //-- one vertex is equal to the isosurface and there are 2 intersections
        else if ( (num = 2) and (noEqual = 1) ) then
        begin
          pt1.SetCoord(inter[0][0], inter[0][1], inter[0][2]);
          pt2.SetCoord(inter[1][0], inter[1][1], inter[1][2]);
          Intersect_3(pt1, pt2, equal[1], lstTr, te, data);
        end;
      end;
      num := 0;
    end;
  end;
  pt1.Free;
  pt2.Free;
  pt3.Free;
  pt4.Free;
  Result := lstTr;
end;


procedure TIsosurface.Intersect_3(pt1, pt2, pt3: TPoint; lstTr: TList;
    te: TTetra; data: double);
var
  outPt: TPoint;
  n: TVector;
begin
  //-- find the "out" point
  outPt := FindOverData(te, data);

  n := TVector.Create;
  TGeomTools.NormalOppositePt(pt1, pt2, pt3, outPt, n);
  lstTr.Add(TTriangle.Create(pt1, pt2, pt3, n));
  n.Free;
end;


procedure TIsosurface.Intersect_4(pt1, pt2, pt3, pt4: TPoint; lstTr: TList;
    te: TTetra; data: double);
var
  outPt: TPoint;
  n: TVector;
begin
  //-- find the "out" point
  outPt := FindOverData(te, data);

  //-- create the a-b-c triangle first
  n := TVector.Create;
  TGeomTools.NormalOppositePt(pt1, pt2, pt3, outPt, n);
  lstTr.Add(TTriangle.Create(pt1, pt2, pt3, n));

  if (TGeomTools.Orient3D(pt1, pt2, outPt, pt3) <>
      TGeomTools.Orient3D(pt1, pt2, outPt, pt4)) then
  begin
    TGeomTools.NormalOppositePt(pt1, pt2, pt4, outPt, n);
    lstTr.Add(TTriangle.Create(pt1, pt2, pt4, n));
  end
  else if (TGeomTools.Orient3D(pt1, pt3, outPt, pt2) <>
           TGeomTools.Orient3D(pt1, pt3, outPt, pt4)) then
  begin
    TGeomTools.NormalOppositePt(pt1, pt3, pt4, outPt, n);
    lstTr.Add(TTriangle.Create(pt1, pt3, pt4, n));
  end
  else  
  begin
    TGeomTools.NormalOppositePt(pt2, pt3, pt4, outPt, n);
    lstTr.Add(TTriangle.Create(pt2, pt3, pt4, n));
  end;
  n.Free;
end;


procedure TIsosurface.Intersect_3_equal(pt1, pt2, pt3: TPoint;
  lstTr: TList; te: TTetra; data: double);
var
  outPt: TPoint;
  n: TVector;
begin
  outPt := FindOverData(te, data);
  if (outPt <> nil) then
  begin
    Intersect_3(pt1, pt2, pt3, lstTr, te, data);
  end
  else
  begin
    outPt := FindUnderData(te, data);
    n := TVector.Create;
    TGeomTools.NormalDirectionPt(pt1, pt2, pt3, outPt, n);
    lstTr.Add(TTriangle.Create(pt1, pt2, pt3, n));
    n.Free;
  end;
end;


//******************************************************************************
// Description: Find the data point of the tetra that value is above the data
//              of the isosurface
//
// Input:       te: the tetra
//              data: the value of the surface to draw
//
// Output:      a pointer to the point if it exists
//              if not, nil is returned
//******************************************************************************
function TIsosurface.FindOverData(te: TTetra; data: double): TPoint;
var
  i: integer;
  outPt: TPoint;
begin
  outPt := nil;
  for i := 1 to 4 do
  begin
    if (te.GetP(i).data > data) then
    begin
      outPt := te.GetP(i);
    end;
  end;
  Result := outPt;
end;


//******************************************************************************
// Description: Find the data point of the tetra that value is under the data
//              of the isosurface
//
// Input:       te: the tetra
//              data: the value of the surface to draw
//
// Output:      a pointer to the point if it exists
//              if not, nil is returned
//******************************************************************************
function TIsosurface.FindUnderData(te: TTetra; data: double): TPoint;
var
  i: integer;
  outPt: TPoint;
begin
  outPt := nil;
  for i := 1 to 4 do
  begin
    if (te.GetP(i).data < data) then
    begin
      outPt := te.GetP(i);
    end;
  end;
  Result := outPt;
end;


procedure TIsosurface.DrawIsosurface(data: double);
var
  lstTr: TList;
  i: integer;
  startTime, endTime: double;
begin
//  startTime := Form1.StampTime;
//  m_mesh.Traverse;
  lstTr := self.ComputeSurface(data);
//  endTime := Form1.StampTime;
//  ShowMessage('total time = ' + FloatToStr(endTime - startTime));
  m_mesh.SetIsosurface(lstTr);

  //-- free the list of triangles
  for i := 0 to (lstTr.Count - 1) do
  begin
    TTriangle(lstTr[i]).Free;
  end;
  lstTr.Free;
end;



end.

