unit OpenGLCamera;

interface

uses
  OpenGL, OpenGLTypes, SysUtils, Math;

type

  TGLCamera = class
  private
    eye: TPoint3;
    lookAt: TPoint3;
    rightVect, upVect, forwardVect: TVector3;
    viewAngle, aspect, nearDist, farDist: double;
  public
    //was private but it's needed to some manipulators
    procedure SetModelViewMatrix; //tell OpenGL where camera is
    //public declarations
    constructor Create;
    procedure SetPosition(eye, look: TPoint3; up: TVector3); overload;
    procedure SetPosition(eyeX, eyeY, eyeZ, lookX, lookY, lookZ, upX, upY, upZ: double); overload;
    procedure GetPosition(eye, lookAt: TPoint3);
    procedure Roll(angle: double);
    procedure Pitch(angle: double);
    procedure Yaw(angle: double);
    procedure Slide(delU, delV, delN: double);
    procedure SetShape(vAng, asp, nearD, farD: double);
    destructor Destroy; override;
  end;

implementation

{ TCamera }

constructor TGLCamera.Create;
begin
  eye := TPoint3.Create;
  lookAt := TPoint3.Create;
  rightVect := TVector3.Create;
  upVect := TVector3.Create;
  forwardVect := TVector3.Create;
end;

procedure TGLCamera.Pitch(angle: double);
var
  sn: double;
  cs: double;
  oldUp: TVector3;
begin
  cs := Cos(Pi/180 * angle); //calculate cosine after changing degrees to radians
  sn := Sin(Pi/180 * angle); //calculate sine after changing degrees to radians
  //new way
  oldUp := upVect;
  upVect.setVector(cs*oldUp.x - sn*forwardVect.x, cs*oldUp.y - sn*forwardVect.y, cs*oldUp.z - sn*forwardVect.z);
  forwardVect.setVector(sn*oldUp.x + cs*forwardVect.x, sn*oldUp.y + cs*forwardVect.y, sn*oldUp.z + cs*forwardVect.z);
  oldUp.Destroy;
  SetModelViewMatrix; //tell OpenGL about changes
end;

procedure TGLCamera.Yaw(angle: double);
var
  sn: double;
  cs: double;
  prevForward: TVector3;
begin
  cs := Cos(Pi/180 * angle); //calculate cosine after changing degrees to radians
  sn := Sin(Pi/180 * angle); //calculate sine after changing degrees to radians
  //try of new way
  prevForward := forwardVect; //remember old FV
  //recalculate forward and right vectors - rotation about up, so no need to change it
  forwardVect.setVector(cs*prevForward.x - sn*rightVect.x, cs*prevForward.y - sn*rightVect.y, cs*prevForward.z - sn*rightVect.z);
  rightVect.setVector(sn*prevForward.x + cs*rightVect.x, sn*prevForward.y + cs*rightVect.y, sn*prevForward.z + cs*rightVect.z);
  prevForward.Destroy;
  SetModelViewMatrix; //tell OpenGL that cameras position has changed
end;

// Rotation about cameras n-axis (cameras look direction)
procedure TGLCamera.Roll(angle: double);
var
  cs, sn: double;
  prevRight: TVector3;
begin
  cs := Cos(Pi/180 * angle); //calculate cosine after changing degrees to radians
  sn := Sin(Pi/180 * angle); //calculate sine after changing degrees to radians
  prevRight := rightVect; //remember old u
  //recalculate u and v vectors - rotation about n, so no need to change n vector
  rightVect.setVector(cs*prevRight.x - sn*upVect.x, cs*prevRight.y - sn*upVect.y, cs*prevRight.z - sn*upVect.z);
  upVect.setVector(sn*prevRight.x + cs*upVect.x, sn*prevRight.y + cs*upVect.y, sn*prevRight.z + cs*upVect.z);
  prevRight.Destroy;
  SetModelViewMatrix; //tell OpenGL about changes
end;

procedure TGLCamera.SetModelViewMatrix;
begin
  lookAt.x := self.eye.x - forwardVect.x;
  lookAt.y := self.eye.y - forwardVect.y;
  lookAt.z := self.eye.z - forwardVect.z;

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  gluLookAt(self.eye.x, self.eye.y, self.eye.z,
            lookAt.x, lookAt.y, lookAt.z,
            0, 1, 0);
end;

procedure TGLCamera.SetPosition(eye, look: TPoint3; up: TVector3);
begin
  SetPosition(eye.x, eye.y, eye.z, look.x, look.y, look.z, up.x, up.y, up.z);
end;

procedure TGLCamera.SetPosition(eyeX, eyeY, eyeZ, lookX, lookY, lookZ, upX, upY, upZ: double);
var
  crossVect, up: TVector3;
begin
  self.eye.setCoordinates(eyeX, eyeY, eyeZ); //store the given eye position
  self.forwardVect.setVector(eyeX - lookX, eyeY - lookY, eyeZ - lookZ); //calculate n
  up := TVector3.Create(upX, upY, upZ);
  crossVect := up.Cross(forwardVect);
  up.Destroy;
  self.rightVect.setVector(crossVect);  // calculate u as a up X n
  forwardVect.Normalize;  // normalize n and u
  rightVect.Normalize;
  crossVect.Destroy;
  crossVect := forwardVect.Cross(rightVect);
  upVect.setVector(crossVect);  // calculate u as a n X u
  crossVect.Destroy;
  SetModelViewMatrix; //tell OpenGL that cameras position has changed
end;

procedure TGLCamera.SetShape(vAng, asp, nearD, farD: double);
begin
  self.viewAngle := vAng;
  self.aspect := asp;
  self.nearDist := nearD;
  self.farDist := farD;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(vAng, asp, nearD, farD);
end;

procedure TGLCamera.Slide(delU, delV, delN: double);
begin
  eye.x := eye.x + delU*rightVect.x + delV*upVect.x + delN*forwardVect.x;
  eye.y := eye.y + delU*rightVect.y + delV*upVect.y + delN*forwardVect.y;
  eye.z := eye.z + delU*rightVect.z + delV*upVect.z + delN*forwardVect.z;
  SetModelViewMatrix;
end;

procedure TGLCamera.GetPosition(eye, lookAt: TPoint3);
begin
  if eye <> nil then
    eye.setCoordinates(self.eye);
  if lookAt <> nil then
    lookAt.setCoordinates(self.lookAt);
end;

destructor TGLCamera.Destroy;
begin
  self.eye.Destroy;
  self.lookAt.Destroy;
  self.rightVect.Destroy;
  self.upVect.Destroy;
  self.forwardVect.Destroy;
  inherited;
end;

end.
